# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Home of the Flatten layer."""

import numpy as np
from .layer import Layer


class Flatten(Layer):
    """Flattens the input while preserving the batch dimension.

    This utility layer is responsible for flattening and reducing to one dimension the internal matrix
    of the neural network to later on process a simpler output vector. The main purpose of this layer is to connect
    the last convolutional layer to the fully connected layer. Note that this does not affect the batch size.

    Args:
        name (str): Name of the layer.
    """

    def __init__(self, name: str = "Flatten"):
        super(Flatten, self).__init__(name=name, has_weights=False)

    def initialize(self, input_shape: tuple):
        """
        Initialize the layer. This method is called by the model when the model is being compiled.

        Args:
            input_shape: The input shape.
        """
        self.input_shape = input_shape
        self.output_shape = (input_shape[0], np.prod(input_shape[1:]))
        self.initialized = True

    def forward(self, x: np.ndarray, training: bool = True) -> np.ndarray:
        """
        Forward pass of the layer.

        Args:
            x: Input to the layer.
            training: For compatibility with the base class.

        Returns:
            A 1D vector.
        """
        self.outputs = np.reshape(x, self.output_shape)
        return self.outputs

    def get_d_inputs(self, delta: np.ndarray) -> np.ndarray:
        """Returns the gradient of the loss with respect to the inputs of the layer.

        Args:
            delta: Delta of the loss with respect to the outputs of the layer.
        Returns:
            The corresponding delta of the layer (d_cost/d_z).
        """
        return np.reshape(delta, self.input_shape)

    def summary(self) -> str:
        return f"{self.name} ({self.input_shape} -> {self.output_shape})"

    def set_weights(self, weights: np.ndarray = None, bias: np.ndarray = None):
        raise NotImplementedError("Flatten layer has no weights")

    def update(self, optimizer, delta: np.ndarray):
        pass

    def count_params(self) -> int:
        return 0



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains mean absolute error loss function."""

import numpy as np
from .loss_function import LossFunction


class MAE(LossFunction):
    """Class that calculates the Mean Absolute Error loss.
    """
    def __init__(self):
        super(MAE, self).__init__(name="mae")

    @staticmethod
    def compute_loss(y_true: np.ndarray, y_pred: np.ndarray):
        """Calculates the Mean Absolute Error loss

        Args:
            y_true: expected output
            y_pred: predictions
        Returns:
            The mean absolute error of the batch.
        """
        return np.abs(y_true - y_pred).mean()

    @staticmethod
    def gradient(y_true: np.ndarray, y_pred: np.ndarray):
        """Calculates the gradient of the Mean Absolute Error loss with respect to the predictions
        
        Args:
            y_true: labels
            y_pred: predictions

        Returns:
            The gradient of the loss with respect to the predictions.
        """
        return np.sign(y_pred - y_true)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains Winograd utils functions. The function of this file is to show an implementation of some functions
    rather than to provide a full implementation of the Winograd algorithm."""

# Source: Fast Algorithms for Convolutional Neural Networks
# By: Andrew Lavin and Scott Gray Nervana Systems
# (https://openaccess.thecvf.com/content_cvpr_2016/papers/Lavin_Fast_Algorithms_for_CVPR_2016_paper.pdf)

# Appendices
# (https://openaccess.thecvf.com/content_cvpr_2016/supplemental/Lavin_Fast_Algorithms_for_2016_CVPR_supplemental.pdf)

import numpy as np
from sympy import *



x, d, g = symbols('image,d,g')


def get_list_m(m: int, r: int):
    """
    This function generates negative and positive numbers in ascend order used for the equation m(image).
        m(image) = image(image-1)(image+1)(image-2)...(image-∞)

    Args:
        m (int): The size of the output
        r (int): The size of the filter

     Returns:
         A list with the numbers chosen to create the polynomial m(image).
         [-1,1,-2,...]
    """
    m_values = [0]
    i = 1
    while len(m_values) < (m + r - 1) - 1:

        if len(m_values) % 2 == 0:
            m_values.append(i)
            i += 1

        elif len(m_values) % 2 == 1:
            m_values.append(-i)

    return m_values


def get_gx_polynomial(r: int):
    """
    This function generates a polynomial g(image) of degree (r-1).

    Args:
        r (int): The size of the filter

     Returns:
         A matrix of dimension (1,r) with each term in each position.
    """

    gx = zeros(0)
    for i in range(r):
        gx = gx.row_insert(len(gx), Matrix([g * x ** i]))
    return gx.T


def get_dx_polynomial(m: int):
    """
    This function generates a polynomial d(image) of degree (m-1).

    Args:
        m (int): The size of the output

     Returns:
         A matrix of dimension (1,d) with each term in each position.
    """
    dx = zeros(0)
    for i in range(m):
        dx = dx.row_insert(len(dx), Matrix([d * x ** i]))
    return dx.T


def get_a(m: int, r: int):
    """
    This function generates a linear system calculating d(i)(image) = d(image) mod m(i) for each i = 0, 1, _, m, len(m_values).
        d(0)(image) = d(image) mod m(0) = d0
        d(1)(image) = d(image) mod m(1) = d0 + d1 + ... + d(m-1)
        .
        .
        d(i)(image) = d(image) mod m(i) = d(i)
        d(len(m_values))(image) = d(image) mod m(len(m_values)) = d(image)[-1]

    Args:
        m (int): The size of the output
        r (int): The size of the filter

     Returns:
         The transformation of the linear system into a matrix of dimension (m+r-1, m).
    """
    A = zeros(0)
    m_values = get_list_m(m, r)
    dx = get_dx_polynomial(m)

    for i in m_values:
        A = A.row_insert(len(A), Matrix(dx.subs(x, i * -1)))

    biggest_coeff = [dx[-1]]
    A_new_row = zeros(len(dx) - 1, 1).row_insert(len(dx), Matrix(biggest_coeff)).T
    A = A.row_insert(len(A) - 1, A_new_row).subs(((x, 1), (d, 1)))
    return A


def get_mx_polynomial(m: int, r: int):
    """
    This function takes the list of numbers from 'get_list_m' and generates que equation for m(image).
        m(image) = image(image-1)(image+1)(image-2)...(image-∞)

    Args:
        m (int): The size of the output
        r (int): The size of the filter

     Returns:
         mx_ply: A representation of m(image) as a polynomial.
         mx_matrix: A representation of m(image) as a matrix which each term in each position.
    """
    mx_matrix = zeros(0)
    mx_ply = 1
    m_values = get_list_m(m, r)

    for i in m_values:
        mx_matrix = mx_matrix.row_insert(len(mx_matrix), Matrix([x + i]))
        mx_ply = (x + i) * mx_ply
    return mx_ply, mx_matrix


def get_b(m: int, r: int):
    """
    This function generates a linear system composed with the calculating M(i)(image) = m(image)/m(i)(image)
    for each i = 0, 1, _, m and the polynomial m(image).
        M(0)(image) = m(image)/m(0)(image)
        M(1)(image) = m(image)/m(1)(image)
        .
        .
        M(i)(image) = m(image)/m(i)(image)
        .
        m(image) = image(image-1)(image+1)(image-2)...


    Args:
        m (int): The size of the output
        r (int): The size of the filter

     Returns:
         The transformation of the linear system into a square matrix of order (m+r-1).
    """
    B = zeros(0)
    m_values = get_list_m(m, r)
    mx, t = get_mx_polynomial(m, r)

    for i in m_values:
        M = Poly(mx / (x + i))
        M_coeffs = M.all_coeffs()[::-1]
        M_coeffs.append(0)
        B = B.row_insert(len(B), Matrix(M_coeffs).T)
    B_last_row = Poly(expand(mx)).all_coeffs()[::-1]
    B = B.T.col_insert(len(B), Matrix(B_last_row))
    return B


def get_m_polynomial(m: int, r: int):
    """
    This function generates a linear system composed with the calculating
    M(i)(image) = m(image)/m(i)(image) for each i = 0, 1, _, m and the polynomial m(image).
        M(0)(image) = m(image)/m(0)(image)
        M(1)(image) = m(image)/m(1)(image)
        .
        .
        M(i)(image) = m(image)/m(i)(image)
        .
        m(image) = image(image-1)(image+1)(image-2)...

    Args:
        m (int): The size of the output
        r (int): The size of the filter

     Returns:
         A matrix of dimension (m+r-1, 1) with each polynomial.
    """
    M = zeros(0)
    m_values = get_list_m(m, r)

    mx, _ = get_mx_polynomial(m, r)

    for i in m_values:
        M = M.row_insert(len(M), Matrix([mx / (x + i)]))
    return M


# n(i)(image)*m(i)(image) + N(i)(image)*M(i)(image) = 1         a*m(i)(image) + b*M(i)(image) = c
def get_n_values(m: int, r: int):
    """
    This function applies the Chinese Remainder Theorem.

    Args:
        m (int): The size of the output
        r (int): The size of the filter

     Returns:
         A matrix of dimension (m+r-1, 1) with value.
    """
    n = zeros(0)
    _, mx = get_mx_polynomial(m, r)
    M = get_m_polynomial(m, r)

    for i in range(len(mx)):
        value = 1 / rem(M[i], mx[i])
        n = n.row_insert(len(n), Matrix([value]))
    return n


def get_gxi(m: int, r: int):
    """
    Description:
        This function generates a linear system calculating g(i)(image) = g(image) mod m(i)
        for each i = 0, 1, _, m, len(m_values).
            g(0)(image) = g(image) mod m(0) = g0
            g(1)(image) = g(image) mod m(1) = g0 + g1 + ... + g(m-1)
            .
            .
            g(i)(image) = g(image) mod m(i) = ....
            g(len(m_values))(image) = g(image) mod m(len(m_values)) = g(image)[-1]

    Args:
        m (int): The size of the output
        r (int): The size of the filter

     Returns:
         The transformation of the linear system into a matrix of dimension (m+r-1, r).
    """
    gxi = zeros(0)
    m_values = get_list_m(m, r)
    mx, _ = get_mx_polynomial(m, r)
    gx = get_gx_polynomial(r)

    for i in m_values:
        gxi = gxi.row_insert(len(gxi), Matrix(gx.subs(x, i * -1)))
    biggest_coeff = [gx[len(gx) - 1]]
    gxi_new_row = zeros(len(gx) - 1, 1).row_insert(len(gx), Matrix(biggest_coeff)).T
    gxi = gxi.row_insert(len(gxi) - 1, gxi_new_row).subs(((x, 1), (g, 1)))
    return gxi


def get_g(m: int, r: int):
    """
    Description:
        This function calculates n.T @ gxi.

    Args:
        m (int): The size of the output
        r (int): The size of the filter

     Returns:
         The resulting matrix of the multiplication which is of dimension (m+r-1, r).
    """
    _g = zeros(0)
    gxi = get_gxi(m, r)
    n = get_n_values(m, r)

    for i in range(len(n)):
        _g = _g.row_insert(len(_g), Matrix(gxi[i, :] * n[i]))
    _g = _g.row_insert(len(_g), gxi.row(len(n)))
    return _g


def winograd_algorithm(image: np.ndarray, m: int, r: int):  # m = number of outputs    r = size of the filter
    """
    Description:
        This function calculates the convolution of the given image.

    Args:
        image (np.ndarray) : The kernel of the given image.
        m (int): The size of the output
        r (int): The size of the filter

     Returns:
         The resulting convolution.
    """

    A = get_a(m, r)
    At = A.T
    B = get_b(m, r)
    Bt = B.T
    G = get_g(m, r)
    gt = G.T

    # To make it easier
    image = np.pad(image, pad_width=1, mode='constant', constant_values=0)

    _g = np.ones((r, r))
    input_size = m + r - 1

    number_of_inputs = image.shape[0] - input_size + 1
    result = np.zeros(1)

    # A minimal 1d algorithm is nested with itself to obtain a minimal 2d algorithm. F (m image m, r image r)
    for i in range(image.shape[0] - input_size + 1):
        for j in range(image.shape[1] - input_size + 1):
            _d = image[i:i + input_size, j:j + input_size]

            # 4x3 3x3 = 4x3 3x4 = 4x4
            s11 = np.dot(G, _g)
            s12 = np.dot(s11, gt)

            # 4x4 4x4 = 4x4 4x4 = 4x4
            s21 = np.dot(B, _d)
            s22 = np.dot(s21, Bt)

            s3 = np.multiply(s12, s22)
            s4 = np.dot(At, s3)
            Y = np.dot(s4, A)
            convolved = Y.sum()

            result = np.hstack([result, convolved])

    result = Matrix(np.delete(result, 0, axis=0).reshape((number_of_inputs, number_of_inputs)))
    return result


if __name__ == "__main__":
    img = np.ones((7, 7), dtype=int)
    winograd_algorithm(img, 3, 3)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains functions related to the Activation Functions."""

from typing import Union
from . import ActivationFunction, Sigmoid, ReLU, Softmax, Linear, Tanh


def get_activation_function(activation_function_name: Union[str, None]) -> Union[ActivationFunction, None]:
    """Returns an activation function object or None if there is no activation function (the same as using 'Linear')

    Args:
        activation_function_name: The name of the activation function. The possible names are:
            - "relu": ReLU
            - "softmax": Softmax
            - "sigmoid": Sigmoid
            - "tanh": Tanh
            - "linear": Linear
            - None: No activation function (i.e. the input is passed through, the same as Linear)
    """
    if activation_function_name is None:
        return None

    if activation_function_name == 'relu':
        return ReLU()
    elif activation_function_name == 'softmax':
        return Softmax()
    elif activation_function_name == 'sigmoid':
        return Sigmoid()
    elif activation_function_name == 'linear':
        return Linear()
    elif activation_function_name == 'tanh':
        return Tanh()
    else:
        raise ValueError(f'Unknown activation function name: {activation_function_name}')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the SimpleConvolutioner class."""

import numpy as np
from typing import Union

from dlfs.convolutioners import Convolutioner


class SimpleConvolutioner(Convolutioner):
    """
    A simple convolutioner that performs a convolution on a single image.

    This convolutioner can be used by Conv2D layers. However, it is not intended to be
    an efficient implementation.

    Usage:
        img = np.array([[2, 2, 1, 3],
                        [0, 3, 2, 1],
                        [1, 1, 0, 2],
                        [2, 0, 0, 1]])

        k = np.array([[1,  0],
                      [2, -2]])
        conv = SimpleConvolutioner(img.shape, k.shape)
        print(conv.convolve(img, k, using_batches=False))

    Args:
        image_size (tuple[int, int] or int): The size of the image to convolve. If an int is provided,
            it is assumed to be the size of the image along the first two dimensions.
        kernel_size (tuple[int, int] or int): The size of the kernel to convolve. If an int is provided,
            it is assumed to be the size of the kernel along the first two dimensions.
        padding (tuple[int, int] or int): The padding to apply to the image. If an int is provided,
            it is assumed to be the padding along the first two dimensions.
        stride (tuple[int, int] or int): The stride of the convolution. If an int is provided,
            it is assumed to be the stride along the first two dimensions.
        data_format (str): The data format of the image and kernel. Must be either 'channels_first' or 'channels_last'.
    """

    def __init__(self,
                 image_size: Union[int, tuple],
                 kernel_size: Union[int, tuple],
                 padding: Union[int, tuple] = (0, 0),
                 stride: Union[int, tuple] = (1, 1),
                 data_format: str = 'channels_last'):

        super().__init__(image_size, kernel_size, padding, stride, data_format=data_format)

    @staticmethod
    def convolve_grayscale(image: np.ndarray,
                           kernel: np.ndarray,
                           stride: Union[int, tuple] = (0, 0),
                           data_format: str = 'channels_last',
                           **kwargs) -> np.ndarray:
        """Performs a valid convolution on a grayscale image with a 2D kernel.

        Args:
            image: A grayscale image.
            kernel: A kernel.
            stride: convolution stride size.
            data_format: The data format of the image. Either 'channels_last' or 'channels_first'.

        Returns:
            A grayscale image.
        """

        # Get the dimensions of the image and kernel
        image_height, image_width = image.shape
        kernel_height, kernel_width = kernel.shape

        # Create the output image
        stride = (stride, stride) if isinstance(stride, int) else stride
        output_height = (image_height - kernel_height) // stride[0] + 1
        output_width = (image_width - kernel_width) // stride[1] + 1
        convolved_image = np.zeros((output_height, output_width))

        # Perform the convolution
        for i in range(output_height):
            for j in range(output_width):
                convolved_image[i, j] = np.sum(image[i * stride[0]:i * stride[0] + kernel_height,
                                                     j * stride[1]:j * stride[1] + kernel_width] * kernel)
        return convolved_image

    @staticmethod
    def convolve_multichannel(image: np.ndarray,
                              kernel: np.ndarray,
                              stride: Union[int, tuple] = (1, 1),
                              data_format: str = 'channels_last',
                              **kwargs) -> np.ndarray:
        """Performs a valid convolution on an image (with multiple channels) with a 3D kernel.

        Args:
            image: A multichannel image.
            kernel: A kernel.
            stride: convolution stride size.
            data_format: The data format of the image. Either 'channels_last' or 'channels_first'.

        Returns:
            A multichannel image.
        """

        # convert image and kernel to channel first if necessary
        if data_format == 'channels_last':
            image = np.moveaxis(image, -1, 0)
            kernel = np.moveaxis(kernel, -1, 0)

        # Get the dimensions of the image and kernel
        if len(kernel.shape) != 3:
            error = 0
            error += 1

        image_channels, image_height, image_width = image.shape
        kernel_channels, kernel_height, kernel_width = kernel.shape

        if kernel_channels != image_channels:
            raise ValueError('The number of channels in the image and kernel must be the same.'
                             f'Image channels: {image_channels}, kernel channels: {kernel_channels}')

        # Create the output image
        stride = (stride, stride) if isinstance(stride, int) else stride
        output_height = (image_height - kernel_height) // stride[0] + 1
        output_width = (image_width - kernel_width) // stride[1] + 1
        convolved_image = np.zeros((output_height, output_width))

        for y in range(output_height):
            for x in range(output_width):
                convolved_image[y][x] = sum(
                    [np.sum(im[y * stride[0]:y * stride[0] + kernel_height, x * stride[1]:x * stride[1] + kernel_width]
                            * kern) for im, kern in zip(image, kernel)])

        # convert image to channel last if necessary
        if data_format == 'channels_last':
            convolved_image = np.moveaxis(convolved_image, 0, -1)

        return convolved_image




# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the stochastic gradient descent optimizer."""

import numpy as np
from typing import Tuple
from dlfs.optimizers import Optimizer
from dlfs.layers import Layer


class SGD(Optimizer):
    """Mini-batch stochastic gradient descent.
    """

    def __init__(self, learning_rate: float = 0.01):
        super(SGD, self).__init__(learning_rate=learning_rate)

    def update(self, layer: Layer, gradients: Tuple[np.ndarray, np.ndarray]):
        """Updates parameters using SGD

        Args:
            layer: The layer to update.
            gradients: delta of parameters (weights and bias)
        """
        dw, db = gradients

        layer.weights -= self.learning_rate * dw
        layer.bias -= self.learning_rate * db

    def add_slot(self, layer: Layer):
        """
        Add a slot for the layer.
        """
        pass



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Home of the Layer base class."""

from abc import ABC, abstractmethod
import numpy as np
from typing import Optional
from dlfs.activation_functions import ActivationFunction, get_activation_function


class Layer(ABC):
    """Base class for all layers.

    Every layer should inherit from this class and implement the abstract methods. A layer is a part of a neural network
    that performs some computation on the input data and produces an output. The output of a layer is the input of the
    next layer. Some layers have weights and biases, which are used to compute the output. These parameters are
    updated during training using backpropagation. This is the reason why there are methods to compute the gradient of
    the loss function with respect to the input of the layer, the weights and the biases. The gradient of the loss
    with respect to the weights and biases is computed in the `update` method.

    Args:
        input_shape (tuple or None): Shape of the input.
        output_shape (tuple): Shape of the output.
        activation (str): Activation function to use.
        name (str): Name of the layer.
        trainable (bool): Whether the layer is trainable.
        has_weights (bool): Whether the layer has weights.

    Attributes:
        input_shape (tuple or None): Shape of the input.
        output_shape (tuple): Shape of the output.
        name (str): Name of the layer. The name is used to identify the layer in the network
            and is used by the optimizer to identify the layer's parameters.
        weights (np.ndarray): Weights of the layer (None by default). If the layer has weights,
            they can be set using the set_weights method or in the initialize method.
        bias (np.ndarray): Bias of the layer (None by default). They are set in the
            initialize method or in the set_weights method.
        has_weights (bool): Whether the layer has weights.
        trainable (bool): Whether the layer is trainable.
        initialized (bool): Whether the layer has been initialized.
        activation (ActivationFunction): Activation function to use.
        inputs (np.ndarray): Inputs of the layer. Needed in the backward pass.
        outputs (np.ndarray): Outputs of the layer before the activation. Needed in the backward pass.
    """

    def __init__(self,
                 input_shape: Optional[tuple] = None,
                 output_shape: Optional[tuple] = None,
                 activation: Optional[str] = None,
                 name: str = None,
                 trainable: bool = True,
                 has_weights: bool = True):

        self.input_shape = input_shape if input_shape else None
        self.output_shape = output_shape
        self.name = name
        self.weights = None
        self.bias = None
        self.__has_weights = has_weights  # Whether the layer has weights and biases.
        self.trainable = trainable
        self.initialized = False
        self.__activation = get_activation_function(activation)
        self.inputs = None
        self.outputs = None  # output of the layer before activation

    # Getters
    # -------------------------------------------------------------------------

    @property
    def activation(self) -> ActivationFunction:
        return self.__activation

    @property
    def has_weights(self) -> bool:
        return self.__has_weights

    # Setters
    # -------------------------------------------------------------------------

    @activation.setter
    def activation(self, activation: str):
        self.__activation = get_activation_function(activation)

    # Abstract methods
    # -------------------------------------------------------------------------

    @abstractmethod
    def initialize(self, input_shape: tuple):
        raise NotImplementedError

    @abstractmethod
    def forward(self, inputs: np.ndarray, training: bool = False) -> np.ndarray:
        """
        Forward pass of the layer.
        Args:
            inputs: input to the layer.
            training: whether the layer is in training mode.
        Returns:
            output of the layer.
        """

    @abstractmethod
    def set_weights(self, weights: np.ndarray = None, bias: np.ndarray = None):
        raise NotImplementedError

    def get_delta(self, d_out: np.ndarray) -> np.ndarray:
        """
        Args:
            d_out: derivative of the cost function with respect to the output of this layer.
        Returns:
            The delta of the layer (d_C/d_z).
        """
        # check if the layer is initialized
        if not self.initialized:
            raise ValueError("The layer is not initialized")

        if self.activation is None:
            delta = d_out
        else:
            activation_gradient = self.activation.gradient(self.outputs)
            # check if the gradient is a matrix of tensors
            #
            # For example, in the case of Dense layers, we have to check if the gradient is a matrix of tensors
            # with shape=(m, n_neurons) or a tensor of jacobian matrices with shape=(m, n_neurons, n_neurons)
            # This is because of activation functions such
            # as softmax which returns a matrix of jacobian matrices.

            if activation_gradient.shape == self.outputs.shape:
                delta = d_out * activation_gradient
            else:
                d_out = d_out[:, np.newaxis, :]
                delta = np.einsum('ijk,ikl->il', d_out, activation_gradient)

                # The above einsum is equivalent (but faster) to the following code:
                # delta = np.empty_like(self.outputs)
                # batch_size = self.input_shape.shape[0]
                # for i in range(batch_size):  # for each sample
                #     delta[i] = d_out[i] @ activation_gradient[i]

        return delta

    @abstractmethod
    def get_d_inputs(self, delta: np.ndarray) -> np.ndarray:
        """
        Returns the derivative of the cost function with respect to the input of the layer.
        Args:
            delta: derivative of the cost function with respect to the output of the layer.
        Returns:
            derivative of the cost function with respect to the input of the layer.
        """

    @abstractmethod
    def summary(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def update(self, optimizer, delta: np.ndarray):
        """Updates the weights and biases of the layer.

        If the layer has no weights, this method does nothing. If the layer has weights,
        the weights and biases are updated according to the optimizer after computing the
        gradient of the cost function with respect to the weights and biases.

        Args:
            optimizer (Optimizer): optimizer to use.
            delta (np.ndarray): gradient of the cost function with respect to the output of the layer (without
                the activation function).
        """
        raise NotImplementedError

    @abstractmethod
    def count_params(self) -> int:
        """Returns the number of parameters of the layer.

        If the layer has no weights, this method returns 0.
        """

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    def __call__(self, inputs, *args, **kwargs):
        return self.forward(inputs, *args, **kwargs)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the convolution utils functions."""

from typing import Union, Tuple
from . import Convolutioner, SimpleConvolutioner, WinogradConvolutioner, PatchConvolutioner


def get_convolution(convolution_type: str,
                    image_size: Union[int, tuple],
                    kernel_size: Union[int, tuple],
                    padding: tuple = (0, 0),
                    stride: Union[int, tuple] = (1, 1),
                    blocksize: Tuple[int, int] = None,
                    data_format: str = "channels_last") -> Convolutioner:
    """
    Returns an activation function object based on the name of the activation function.
    Args:
        convolution_type: The name of the activation function. The possible names are:
            - "simple": Simple convolution.
            - "winograd": Winograd convolution.
        image_size: The size of the matrix.
        kernel_size: The size of the kernel.
        padding: Whether to use padding or not.
        stride: The stride of the convolution.
        blocksize (Tuple[int, int]): the size of the block, only used with Winograd.
        data_format (str): The data format of the input.

    Returns:
        An activation function object or None if there is no activation function (the same as using 'Linear').
    """

    if convolution_type == "simple":
        return SimpleConvolutioner(image_size, kernel_size, padding, stride, data_format=data_format)
    elif convolution_type == "winograd":
        return WinogradConvolutioner(image_size, kernel_size, padding, stride, blocksize, data_format=data_format)
    elif convolution_type == "patch":
        return PatchConvolutioner(image_size, kernel_size, padding, stride, data_format=data_format)
    else:
        raise ValueError(f'Unknown activation function name: {convolution_type}')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Home of the Loss base class."""

from abc import ABC, abstractmethod
from numpy import ndarray


class LossFunction(ABC):
    """Base class for loss functions.

    Loss functions are used to compute the loss between the prediction and the
    target. They can also be used as metrics.
    """

    def __init__(self, name):
        self.name = name

    @staticmethod
    @abstractmethod
    def compute_loss(y_true, y_pred) -> float:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def gradient(y_true, y_pred) -> ndarray:
        raise NotImplementedError

    def __call__(self, y_true, y_pred):
        return self.compute_loss(y_true, y_pred)



from abc import ABC, abstractmethod
import numpy as np
from typing import Tuple

from dlfs.layers import Layer


class Optimizer(ABC):
    """Base class for all optimizers.

    Args:
        learning_rate (float): learning rate for the optimizer.
    """

    def __init__(self, learning_rate):
        self.learning_rate = learning_rate

    @abstractmethod
    def update(self, layer: Layer, gradients: Tuple[np.ndarray, np.ndarray]):
        """Updates parameters based on delta.
        Args:
            layer: The layer to update.
            gradients: Gradients of the parameters passed as a tuple of two numpy arrays composed of the delta of
                the weights and biases.
        """
        raise NotImplementedError

    @abstractmethod
    def add_slot(self, layer: Layer):
        """Adds a slot to the optimizer.

        Args:
            layer: The layer to add a slot to.
        """
        raise NotImplementedError



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains BinaryCrossentropy class."""

import numpy as np
from .loss_function import LossFunction


class BinaryCrossentropy(LossFunction):
    """Binary cross-entropy loss function

    A loss used for classify binary categories. This loss function is commonly used with one terminal neuron and a
    sigmoid function.
    """

    def __init__(self, name='binary_crossentropy'):
        super().__init__(name)

    @staticmethod
    def compute_loss(y_true, y_pred):
        """Compute the binary cross entropy loss."""
        y_pred_clipped = np.clip(y_pred, 1e-7, 1 - 1e-7)
        return np.mean(-y_true * np.log(y_pred_clipped) - (1 - y_true) * np.log(1 - y_pred_clipped))

    @staticmethod
    def gradient(y_true, y_pred):
        """Compute the gradient of the binary cross entropy loss."""
        return y_pred - y_true



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the stochastic gradient descent momentum optimizer."""

import numpy as np
from typing import Tuple
from .optimizer import Optimizer
from dlfs.layers import Layer


class SGDMomentum(Optimizer):

    def __init__(self, learning_rate: float = 0.01, momentum: float = 0.9):
        super(SGDMomentum, self).__init__(learning_rate)
        self.momentum = momentum
        self.__v = {}

    def update(self, layer: Layer, gradients: Tuple[np.ndarray, np.ndarray]):
        """Updates the weights of the layer using the delta.

        Args:
            layer: The layer to update.
            gradients: The delta of the layer.
        """
        dw, db = gradients

        vw, vb = self.__v[layer.name]

        # update velocity. It uses exponential weighted average.
        vw = self.momentum * vw + (1 - self.momentum) * dw
        vb = self.momentum * vb + (1 - self.momentum) * db

        # update parameters
        layer.weights -= self.learning_rate * vw
        layer.bias -= self.learning_rate * vb

        # store the velocity
        self.__v[layer.name] = (vw, vb)

    def add_slot(self, layer: Layer):
        """Adds a slot for the layer.

        Args:
            layer: The layer to add a slot for.
        """
        self.__v[layer.name] = (np.zeros_like(layer.weights), np.zeros_like(layer.bias))

    def reset(self):
        """Resets the optimizer."""
        self.__v = {}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the get_metric function."""

from typing import Union
from dlfs.metrics import Metric, Accuracy
from dlfs.losses import *


def get_metric(name: str) -> Union[Metric, LossFunction]:
    """Returns a metric object by name.
    """

    if name == 'accuracy':
        return Accuracy()
    elif name == 'mae':
        return MAE()
    elif name == 'mse':
        return MSE()
    else:
        raise ValueError(f'Unknown metric: {name}')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the hyperbolic tangent (tanh) activation function"""

import numpy as np
from .activation_function import ActivationFunction


class Tanh(ActivationFunction):
    """Hyperbolic tangent activation function.

    The hyperbolic tangent activation function is defined as:
    tanh(x) = (e^x - e^-x) / (e^x + e^-x)
    """

    def __init__(self):
        super().__init__(name='tanh')

    @staticmethod
    def forward(x: np.ndarray) -> np.ndarray:
        """Forward pass of the tanh activation function."""
        # luckily, numpy has a built-in tanh function
        return np.tanh(x)

    @staticmethod
    def gradient(z: np.ndarray) -> np.ndarray:
        """Computes the gradient of the tanh activation function."""
        return 1 - np.square(z)
